#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/net.h>
#include <linux/types.h>
#include <linux/socket.h>
#include <linux/list.h>
#include <net/tcp.h>
#include <net/ip.h>
#include <linux/inet.h>
#include <net/protocol.h>
#include <net/sock.h>
#include "consts.h"
#include <linux/new_capi.h>
#include "debug.h"
#include "appl.h"
#include "capi_pipe.h"
#include "local_capi.h"
#include "zugriff.h"

static struct socket *capi_socket=NULL;

struct _capi_connections {
    struct list_head list;
    struct socket *sock;
    unsigned ApplId;
};

LIST_HEAD(capi_conn_list);


/*-------------------------------------------------------------------------------------*\
\*-------------------------------------------------------------------------------------*/
static int capi_oslib_socket_write(struct socket *sock, unsigned char* buffer, unsigned length, int wait) {
    struct msghdr        msg;
    struct iovec        iov;
    
    mm_segment_t oldfs;

    if (sock->sk == NULL) {
        return 0;
    }

    iov.iov_base = (void *)buffer;
    iov.iov_len = length;

    msg.msg_name =      NULL;
    msg.msg_namelen =   0;
    msg.msg_iov =       &iov;
    msg.msg_iovlen      = 1;
    msg.msg_control     = NULL;
    msg.msg_controllen  = 0;
    msg.msg_flags       = wait ? 0 : MSG_DONTWAIT;

    oldfs = get_fs(); set_fs(KERNEL_DS);
    length = sock_sendmsg(sock, &msg, length);
    set_fs(oldfs);
    DEB_INFO("capi_oslib_socket_write: %d bytes written\n", length);
    return length;
}

/*-------------------------------------------------------------------------------------*\
\*-------------------------------------------------------------------------------------*/
static int capi_oslib_socket_read(struct socket *sock, unsigned char* buffer, unsigned length, int wait)
{
    struct msghdr   msg;
    struct iovec    iov;
    int ret;

    mm_segment_t    oldfs;
    
    if (sock->sk == NULL) {
        return 0;
    }

    iov.iov_base        = (void *)buffer;
    iov.iov_len         = length;
    msg.msg_name        = 0;
    msg.msg_namelen     = 0;
    msg.msg_iov         = &iov;
    msg.msg_iovlen      = 1;
    msg.msg_control     = NULL;
    msg.msg_controllen  = 0;
    msg.msg_flags       = MSG_DONTWAIT;

    oldfs = get_fs(); set_fs(KERNEL_DS);
    ret = sock_recvmsg(sock, &msg, length, wait ? 0 : MSG_DONTWAIT);
    set_fs(oldfs);
    return ret;
}

/*-------------------------------------------------------------------------------------*\
\*-------------------------------------------------------------------------------------*/
static unsigned int capi_oslib_socket_put(unsigned short ApplId, unsigned char* buffer)
{
    unsigned int status = ERR_SendBusy;
    
    while (status == ERR_SendBusy) {
        status = LOCAL_CAPI_PUT_MESSAGE(SOURCE_SOCKET_CAPI, ApplId, buffer);
        DEB_INFO("LOCAL_CAPI_PUT_MESSAGE(%d, %p) -> %d\n", ApplId, buffer, status);
    }
    
    return status;
}
 
/*-------------------------------------------------------------------------------------*\
\*-------------------------------------------------------------------------------------*/
static void capi_oslib_socket_get(void* arg)
{
    unsigned int status = ERR_NoError;
    unsigned char* buffer = NULL;
    struct _capi_connections *conn = (struct _capi_connections*)arg;

    if (conn == NULL) return;
    DEB_INFO("capi_oslib_socket_get for Appl %d\n", conn->ApplId);
    if (conn->ApplId == (unsigned)-1) return;


    while (status == ERR_NoError) {
        status = LOCAL_CAPI_GET_MESSAGE(SOURCE_SOCKET_CAPI, conn->ApplId, &buffer, CAPI_NO_SUSPEND);
        DEB_INFO("LOCAL_CAPI_GET_MESSAGE(%d, CAPI_NO_SUSPEND) -> %04x, %p\n", conn->ApplId, status, buffer);
        if (status == ERR_NoError) {
            struct __attribute__ ((packed)) _capi_message *C = 
                (struct __attribute__ ((packed)) _capi_message *)buffer;
            unsigned written = 0;

            while (written < C->capi_message_header.Length) {
                int err;
                err = capi_oslib_socket_write(conn->sock, &buffer[written], C->capi_message_header.Length - written, 1);
                if (err <= 0) return;
                written += err;
            }

            if (written < C->capi_message_header.Length) return;

            if (CA_IS_DATA_B3_IND(buffer)) {
                unsigned char* data;

                data = C->capi_message_part.data_b3_ind.Data;
                written = 0;

                while (written < C->capi_message_part.data_b3_ind.DataLen) {
                    int err;
                    err = capi_oslib_socket_write(conn->sock, &data[written], C->capi_message_part.data_b3_ind.DataLen - written, 1);
                    if (err <= 0) return;
                    written += err;
                }
            }
        }
        DEB_INFO("get done\n");
    }
}

/*-------------------------------------------------------------------------------------*\
\*-------------------------------------------------------------------------------------*/
static int capi_oslib_socket_conn_thread(void* args)
{
    struct socket* sock = (struct socket*)args;
    unsigned char* buffer = NULL;
    unsigned buflen = 128 + 2048;
    struct __attribute__ ((packed)) _capi_message *C;
    struct _capi_connections conn;
    struct work_struct rx_work;
    
    if (args == NULL) {
        DEB_INFO("capi_oslib_socket_conn_thread called with NULL socket\n");
        return -1;
    }
    DEB_INFO("capi_oslib_socket_conn_thread started...\n");
    conn.ApplId = (unsigned)-1;
    conn.sock = sock;

    buffer = kmalloc(buflen, GFP_KERNEL);
    if (buffer == NULL) goto exit_reader_thread;
    C = (struct __attribute__ ((packed)) _capi_message *)buffer;

    daemonize("capiconn");

    /*-------------------------------------------------------------------------------------*\
     * CAPI Profile etc.pp rberschicken
     *
     * 4 Bytes Num. Controllers
     *      64 Bytes Manufacturer
     *      4  Bytes CAPI Major
     *      4  Bytes CAPI Minor
     *      4  Bytes Manu Major
     *      4  Bytes Manu Minor
     *      8  Bytes Serial No.
     *      64 Bytes Profile
    \*-------------------------------------------------------------------------------------*/
    {
        unsigned char info[64];
        unsigned written;
        unsigned int num_controllers, i;

        CAPI_GET_PROFILE(info, 0);
        num_controllers = EXTRACT_DWORD(info);

        written = 0;
        while (written < 4)
        {
            int err;
            err = capi_oslib_socket_write(sock, &info[written], 4 - written, 1);
            if (err <= 0) break;
            written += err;
        }
        if (written < 4) goto exit_reader_thread;

        for (i = 1; i <= num_controllers; i++) {
            written = 0;
            CAPI_GET_MANUFACTURER(info);
            DEB_INFO("start info controller %d\n", i);
            while (written < 64)
            {
                int err;
                err = capi_oslib_socket_write(sock, &info[written], 64 - written, 1);
                if (err <= 0) break;
                written += err;
            }
            if (written < 64) goto exit_reader_thread;

            CAPI_GET_VERSION((unsigned int*)&info[0], (unsigned int*)&info[4], (unsigned int*)&info[8],
                             (unsigned int*)&info[12]);
            written = 0;
            while (written < 16)
            {
                int err;
                err = capi_oslib_socket_write(sock, &info[written], 16 - written, 1);
                if (err <= 0) break;
                written += err;
            }
            if (written < 16) goto exit_reader_thread;

            CAPI_GET_SERIAL_NUMBER(i, info);
            written = 0;
            while (written < 8)
            {
                int err;
                err = capi_oslib_socket_write(sock, &info[written], 8 - written, 1);
                if (err <= 0) break;
                written += err;
            }
            if (written < 8) goto exit_reader_thread;

            CAPI_GET_PROFILE(info, i);
            written = 0;
            while (written < 64)
            {
                int err;
                err = capi_oslib_socket_write(sock, &info[written], 64 - written, 1);
                if (err <= 0) break;
                written += err;
            }
            if (written < 64) goto exit_reader_thread;
            DEB_INFO("done info controller %d\n", i);
        }
    }

    
    /*-------------------------------------------------------------------------------------*\
    \*-------------------------------------------------------------------------------------*/
    for (;;)
    {
        int err;
        unsigned length_read = 0;
        unsigned msg_len = 0;
        
        /* von socket lesen */
        err = capi_oslib_socket_read(sock, buffer, 2, 1);
        if (err <= 0) {
            break;
        }

        msg_len = buffer[0] | (buffer[1] << 8);
        length_read = 2;

        if ((msg_len < 8) || (msg_len > buflen)) {
            DEB_ERR("capi_oslib_socket_read: Message Length %d looks fishy!\n", msg_len);
            if (msg_len > buflen) {
                msg_len = buflen;
            }
        }

        while (length_read < msg_len) {
            err = capi_oslib_socket_read(sock, &buffer[length_read], msg_len - length_read, 1);
            if (err <= 0) {
                DEB_ERR("err reading msg\n");
                break;
            }
            length_read += err;
        }

        /* fehler? */
        if (err <= 0) break;
        DEB_INFO("capi_oslib: read message of %d bytes len\n", msg_len);

#if 0
        {
            unsigned i, len;
            len = 0;
            while (i + len < msg_len) {
                printk(KERN_INFO);
                for (i = 0; i < 8 && i < msg_len; i++)
                {
                    printk("%02x ", buffer[len + i]);
                }
                printk("\n");
                len += 8;
            }
        }
#endif
        
        switch (buffer[4])
        {
            case 0x86:    /* DATA_B3_REQ */
                if (buffer[5] == 0x80) {
                    struct __attribute__ ((packed)) _capi_message *C = (struct __attribute__ ((packed)) _capi_message *)buffer;
                    unsigned char* data_buffer = NULL;
                    unsigned data_read;

                    data_buffer = LOCAL_CAPI_NEW_DATA_B3_REQ_BUFFER(SOURCE_SOCKET_CAPI, 
                            C->capi_message_header.ApplId, C->capi_message_part.data_b3_req.NCCI);
                    if (data_buffer) {
                        data_read = 0;
                        while (data_read < C->capi_message_part.data_b3_req.DataLen) {
                            int err;
                            err = capi_oslib_socket_read(sock, &data_buffer[data_read], C->capi_message_part.data_b3_req.DataLen - data_read, 1);
                            if (err < 0) break;
                            data_read += err;
                        }
                        if (data_read < C->capi_message_part.data_b3_req.DataLen) break;
                        C->capi_message_part.data_b3_req.Data = data_buffer;
                    }
                }
                capi_oslib_socket_put(C->capi_message_header.ApplId, buffer);
                break;
            case 0xFD:
                {
                    unsigned maxNCCIs;
                    unsigned windowsize;
                    unsigned b3_blocksize;
                    unsigned result;

                    if (conn.ApplId != (unsigned)-1) {
                        list_del(&conn.list);
                        LOCAL_CAPI_SET_NOTIFY(SOURCE_SOCKET_CAPI, conn.ApplId, NULL);
                        LOCAL_CAPI_RELEASE(SOURCE_SOCKET_CAPI, conn.ApplId);
                        conn.ApplId = (unsigned)-1;
                    }
                    maxNCCIs = buffer[8] | (buffer[9] << 8) | (buffer[10] << 16) | (buffer[11] << 24);
                    windowsize = buffer[12] | (buffer[13] << 8) | (buffer[14] << 16) | (buffer[15] << 24);
                    b3_blocksize = buffer[16] | (buffer[17] << 8) | (buffer[18] << 16) | (buffer[19] << 24);
                    result = LOCAL_CAPI_REGISTER(SOURCE_SOCKET_CAPI, 1024 + (1024 * maxNCCIs), maxNCCIs, windowsize, b3_blocksize, &conn.ApplId);
                    DEB_INFO("register result: %d\n", result);
                    if (result == ERR_NoError) {
                        unsigned char response[10];
                        unsigned sent;
                        list_add(&conn.list, &capi_conn_list);
                        
                        /* Antwort mit ApplID schicken */
                        memcpy((void*)response, buffer, 10);
                        response[2] = (conn.ApplId & 0xFF);         /* ApplID einsetzen */
                        response[3] = (conn.ApplId >> 8) & 0xFF;
                        response[8] = 0;
                        response[9] = 0;

                        sent = 0;
                        while (sent < 10) {
                            int err;
                            err = capi_oslib_socket_write(sock, &response[sent], 10 - sent, 1);
                            if (err <= 0) break;
                            sent += err;
                        }
                        if (sent < 10) {
                            DEB_ERR("err sending register result\n");
                            goto exit_reader_thread;
                        }

                        DEB_INFO("register success result sent\n");
                        /* workqueue item erzeugen und einhngen */
                        INIT_WORK(&rx_work, capi_oslib_socket_get, (void*)&conn);
                        result = LOCAL_CAPI_SET_NOTIFY(SOURCE_SOCKET_CAPI, conn.ApplId, &rx_work);
                    } else {
                        unsigned char response[10];
                        unsigned sent;

                        conn.ApplId = (unsigned)-1;
                        /* Fehlerwert zurckschicken */
                        memcpy((void*)response, buffer, 10);
                        response[8] = (result & 0xFF);
                        response[9] = (result >> 8) & 0xFF;

                        sent = 0;
                        while (sent < 10) {
                            int err;
                            err = capi_oslib_socket_write(sock, &response[sent], 10 - sent, 1);
                            if (err <= 0) break;
                            sent += err;
                        }
                        if (sent < 10) {
                            DEB_ERR("err sending register result(2)\n");
                            goto exit_reader_thread;
                        }
                    }
                }
                break;
            default:
                capi_oslib_socket_put(C->capi_message_header.ApplId, buffer);
                break;
        }
    }
exit_reader_thread:
    if (buffer) kfree(buffer);
    sock_release(sock);
    if (conn.ApplId != (unsigned)-1) {
        list_del(&conn.list);
        LOCAL_CAPI_SET_NOTIFY(SOURCE_SOCKET_CAPI, conn.ApplId, NULL);
        LOCAL_CAPI_RELEASE(SOURCE_SOCKET_CAPI, conn.ApplId);
        conn.ApplId = (unsigned)-1;
    }
    DEB_INFO("capi_oslib_socket_conn_thread stopped...\n");
    return 0;
}

/*-------------------------------------------------------------------------------------*\
\*-------------------------------------------------------------------------------------*/
static int capi_oslib_socket_accept_thread(void* args)
{
    daemonize("capisock");

    for (;capi_socket != NULL;)
    {
        struct socket* sock;
        int err;

        if (sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock) < 0)
        {
            DEB_INFO("capi_oslib_socket_accept_thread: socket_alloc failed\n");
            sock_release(capi_socket);
            capi_socket = NULL;
            continue;
        }

        DEB_INFO("capi oslib accept\n");
        err = capi_socket->ops->accept(capi_socket, sock, 0);
        if (err < 0) {
            DEB_INFO("capi oslib accept failed with %d\n", err);
            sock_release(sock);
            continue;
        }

        kernel_thread(capi_oslib_socket_conn_thread, (void*)sock, 0);
    }
    DEB_INFO("capi_oslib_socket_accept_thread() complete_and_exit() ?!\n");
    return 0;
}

/*-------------------------------------------------------------------------------------*\
\*-------------------------------------------------------------------------------------*/
void capi_oslib_socket_init(void)
{
    struct sockaddr_in addr;
    
    if (sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &capi_socket) < 0)
    {
        DEB_INFO("capi_oslib_socket_init(): Error during creation of socket\n");
        capi_socket = NULL;
    }

    addr.sin_family = AF_INET;
    addr.sin_port = htons((unsigned short)1234);
    addr.sin_addr.s_addr = 0;

    if (capi_socket->ops->bind(capi_socket, (struct sockaddr*)&addr, sizeof(addr)) < 0)
    {
        DEB_INFO("capi_oslib_socket_init(): bind failed\n");
        sock_release(capi_socket);
        capi_socket = NULL;
        return;
    }

    if (capi_socket->ops->listen(capi_socket, 1) < 0)
    {
        DEB_INFO("capi_oslib_socket_init(): listen failed\n");
        sock_release(capi_socket);
        capi_socket = NULL;
        return;
    }

    /* Socket erzeugen und listen() erfolgreich, Thread starten fr accept() */
    LOCAL_CAPI_INIT(SOURCE_SOCKET_CAPI);
    kernel_thread(capi_oslib_socket_accept_thread, NULL, 0);
}

/*-------------------------------------------------------------------------------------*\
\*-------------------------------------------------------------------------------------*/
void capi_oslib_socket_release(void)
{
    if (capi_socket) {
        sock_release(capi_socket);
        capi_socket = NULL;
    }
}
